package main

import (
	"fmt"
	"net/http"
	_ "net/http/pprof"
	"os"
	"os/signal"
	"path/filepath"
	"runtime/pprof"
	"sync"
	"syscall"
	"time"

	"github.com/urfave/cli"

	"gitlab.33.cn/link33/sidecar/internal"
	"gitlab.33.cn/link33/sidecar/internal/app"
	"gitlab.33.cn/link33/sidecar/internal/loggers"
	"gitlab.33.cn/link33/sidecar/internal/repo"
	"gitlab.33.cn/link33/sidecar/pkg/log"
	"gitlab.33.cn/link33/sidecar/tool"
)

var startCMD = cli.Command{
	Name:   "start",
	Usage:  "Start a long-running daemon process",
	Action: start,
}

func start(ctx *cli.Context) error {
	fmt.Println(getVersion(true))

	repoRoot, err := repo.PathRootWithDefault(ctx.GlobalString("repo"))
	if err != nil {
		return err
	}

	repo.SetPath(repoRoot)

	config, err := repo.UnmarshalConfig(repoRoot)
	if err != nil {
		return fmt.Errorf("init config error: %s", err)
	}

	err = log.Initialize(
		log.WithReportCaller(config.Log.ReportCaller),
		log.WithPersist(true),
		log.WithFilePath(filepath.Join(repoRoot, config.Log.Dir)),
		log.WithFileName(config.Log.Filename),
		log.WithMaxSize(2*1024*1024),
		log.WithMaxAge(24*time.Hour),
		log.WithRotationTime(24*time.Hour),
	)
	if err != nil {
		return fmt.Errorf("log initialize: %w", err)
	}
	// init loggers map for sidecar
	loggers.InitializeLogger(config)

	var sidecar internal.Launcher
	fmt.Println(repoRoot)
	fmt.Println(config)
	sidecar, err = app.NewSidecar(repoRoot, config)
	if err != nil {
		return err
	}

	httpPProf(config.Port.PProf)
	runtimePProf()

	var wg sync.WaitGroup
	wg.Add(1)
	handleShutdown(sidecar, &wg)

	if err := sidecar.Start(); err != nil {
		return err
	}

	wg.Wait()

	logger.Info("App exits")
	return nil
}

func handleShutdown(sidecar internal.Launcher, wg *sync.WaitGroup) {
	stop := make(chan os.Signal)
	signal.Notify(stop, syscall.SIGTERM)
	signal.Notify(stop, syscall.SIGINT)

	go func() {
		<-stop
		fmt.Println("received interrupt signal, shutting down...")
		if err := sidecar.Stop(); err != nil {
			logger.Error("sidecar stop: ", err)
		}

		wg.Done()
		os.Exit(0)
	}()
}

func httpPProf(port int64) {
	go func() {
		addr := fmt.Sprintf("localhost:%d", port)
		fmt.Printf("Pprof on localhost:%d\n\n", port)
		tool.Asset(http.ListenAndServe(addr, nil))
	}()
}

// runtimePProf will record the cpu or memory profiles every 5 second.
func runtimePProf() {
	tick := time.NewTicker(time.Second * 5)
	defaultPathRoot, err := repo.PathRoot()
	tool.Asset(err)
	rootPath := filepath.Join(defaultPathRoot, "/pprof/")
	exist := tool.Exist(rootPath)
	if !exist {
		err := os.Mkdir(rootPath, os.ModePerm)
		if err != nil {
			fmt.Printf("----- runtimePProf start failed, err: %s -----\n", err.Error())
			return
		}
	}
	go func() {
		for {
			select {
			case <-tick.C:
				pprof.StopCPUProfile()
				subCpuPath := fmt.Sprint("cpu-", time.Now().Format("20060102-15:04:05"))
				cpuPath := filepath.Join(rootPath, subCpuPath)
				cpuFile, _ := os.Create(cpuPath)
				_ = pprof.StartCPUProfile(cpuFile)

				subMemPath := fmt.Sprint("mem-", time.Now().Format("20060102-15:04:05"))
				memPath := filepath.Join(rootPath, subMemPath)
				memFile, _ := os.Create(memPath)
				_ = pprof.WriteHeapProfile(memFile)
				_ = memFile.Close()
			}
		}
	}()
}
